import numpy as np
import scipy.integrate as ig
import astropy.units as u

from .config import DefaultConfig
from .coords import Grid
from .utils import get_key_list, get_dict_from_keys, recursive_subclasses, merge_data
from .custom_exceptions import EngineInputError, WavesetMismatch, DataError, RangeError, DataConfigurationError

def ProfileFactory(config={}, **kwargs):

    """
    Function to take profile configuration data and build/return a configured instance of 
    the appropriate distribution generator.

    Parameters
    ----------
    webapp: bool
        Switch to toggle strict API checking
    config: dict
        Configuration data in engine API dict format
    **kwargs: keyword/value pairs
        Additional configuration data
    """
    #all_config = merge_data(config, dict(**kwargs))
    all_config = config
    if 'geometry' in all_config.shape:
        method = all_config.shape['geometry'].replace('_', '')
    else:
        msg = "Must specify type of profile to create."
        raise EngineInputError(value=msg)
        


    # we need to run the API checks here after merging in the normalization type-specific defaults
    #if webapp:
    #    self._api_checks(all_config)

    types = recursive_subclasses(Distribution)
    methods = [t.__name__.lower().replace('distribution', '') for t in types]
    type_map = dict(list(zip(methods, types)))

    if method not in methods:
        msg = "Unsupported or not yet implemented profile type: %s" % method
        raise EngineInputError(value=msg)
    else:
        cls = type_map[method](config,**kwargs)
        return cls
        
class Distribution(DefaultConfig):
    """
    Abstract class for source profiles. 
    
    Parameters
    ----------
    src: dict
        Details of the source geometry to generate, from the Source block of the input files.

    Attributes
    ----------
    xoff: float
        Offset in X direction
    yoff: float
        Offset in Y direction
    pa: float
        Position angle of major axis
    xrot: np.ndarray
        rotated and shifted indicies, X axis values
    yrot: np.ndarray
        rotated and shifted indicies, Y axis values
    pix_area_sqarcsec: float
        area of a detector pixel in square arcseconds
    prof: np.ndarray
        raw 2D profile in detector units
    norm_prof: np.ndarray
        normalized 2D profile in detector units
        
    Methods
    -------
    sampled()
        return the normalized 2D profile
    raw()
        return the raw 2D profile

    """
    def __init__(self,src):
        self.pa=src.position['orientation']
        self.xoff=src.position['x_offset']
        self.yoff=src.position['y_offset']
        yrot, xrot = src.grid.shift_rotate(self.yoff, self.xoff, self.pa)
        self.xrot = np.abs(xrot)
        self.yrot = np.abs(yrot)
        self.pix_area_sqarcsec = src.grid.xsamp * src.grid.ysamp

    def pixelscale(self):
        if self.surf_area_units in ['sr']:
            arcsec2 = u.arcsec * u.arcsec
            normfactor = self.pix_area_sqarcsec / u.sr.to(arcsec2)  # convert area in steradians to area in pixels
        elif self.surf_area_units in ['arcsec^2', None]: # 'None' should be an option because integrated 
                                                                                                         #  flux technically shouldn't have units (internally, the grid is arcsec)
            normfactor = self.pix_area_sqarcsec
        else:
            msg = "Unsupported surface area unit: %s" % self.surf_area_units
            raise EngineInputError(value=msg)
            
        return normfactor
        
    def normalize(self,src):
        if src.shape['norm_method'] == 'integ_infinity':
                self.integrate_infinity()
        elif src.shape['norm_method'] == 'surf_scale':
            if src.shape['geometry'] == 'flat': # really needs to be refactored into something more clever
                msg = "Normalization method {0:} not supported for profile type {1:}".format(src.shape['norm_method'],src.shape['geometry'])
                raise EngineInputError(value=msg)
            else:
                        self.surface_scale()
        elif src.shape['norm_method'] == 'surf_center':
                self.surface_center()
        else:
            msg = "Normalization method {0:} not supported for profile type {1:}".format(src.shape['norm_method'],src.shape['geometry'])
            raise EngineInputError(value=msg)
    
    def sampled(self):
        return self.norm_prof
        
    def raw(self):
        return self.prof
        
class PointDistribution(Distribution):
    """
    Use pandeia.engine.coords.Grid.point_source() to generate a point source with 
    subpixel positioning.
    
    Parameters
    ----------
    src: dict
        Details of the source geometry to generate, from the Source block of the input files.

    Attributes
    ----------
    xoff: float
        Offset in X direction
    yoff: float
        Offset in Y direction
    prof: np.ndarray
        raw 2D profile in detector units
    norm_prof: np.ndarray
        normalized 2D profile in detector units

    """
    def __init__(self,src):

        self.xoff=src.position['x_offset']
        self.yoff=src.position['y_offset']

        # This function both creates and fills the point source profile. It should
        # eventually be refactored to work the same way as other sources
        self.prof = src.grid.point_source(xoff=self.xoff, yoff=self.yoff)
        
        self.norm_prof = self.prof


class SersicDistribution(Distribution):
    """
    Create a 2-dimensional elliptical source on the current grid. The intensity 
    profile is described by a Sersic profile, I(r) = I(0) * exp(-(r/r_scale)**(1/n)), 
    where r_scale is the scale length where I(r) = I(0)/e and n is the Sersic index. 
    The ellipticity is governed by specifying major and minor axis scale lengths 
    separately.
    
    
    Parameters
    ----------
    src: dict
        Details of the source geometry to generate, from the Source block of the input files.

    Attributes
    ----------
    minor: float
        Minor axis scale length
    major: float
        Major axis scale length
    pa: float
        Position angle in degrees of major axis measured positive in +X direction
    xoff: float
        Offset in X direction
    yoff: float
        Offset in Y direction
    sersic_index: float
        Sersic profile shape parameter. 0.5 => gaussian, 1.0 => exponential, 4.0 => de Vaucouleurs
    prof: np.ndarray
        raw 2D profile in detector units
    norm_prof: np.ndarray
        normalized 2D profile in detector units

    Methods
    -------
        integrate_infinity():
            Normalization is performed using scipy.integrate.dblquad to integrate the profile 
        in 2D from -Inf to +Inf in both axes.  This will account for flux that falls 
        outside of the FOV. This works well for sersic_index <= 2.0 or so, but slow 
        convergence for larger values can impact performance noticeably.  For example, it 
        takes 10 times longer to normalize a de Vaucouleurs profile with sersic_index=4.0 
        than an exponential disc with sersic_index=1.  Truncating the integration range 
        can help, but also significantly impacts accuracy as sersic_index increases.  
        Ultimate solution may be to implement the integration in C somehow.
    surface_center():
        Normalization is to the surface brightness of the center of the profile. This 
        formulation is already normalized at that point.
    surface_scale():
        Normalization is to the surface brightness of a point at the e-folding radius of
        this profile, where intensity is 1/e of the central intensity.
    """
    def __init__(self,src):
        Distribution.__init__(self,src)
        self.major=src.shape['major']
        self.minor=src.shape['minor']
        self.sersic_index=src.shape['sersic_index']
        self.surf_area_units = src.shape['surf_area_units']


        self.prof = self._sersic_func(self.yrot, self.xrot, self.major, self.minor, self.sersic_index)
        
        self.normalize(src)

    def integrate_infinity(self):
        # integrate the Sersic profile to get the total flux for normalization, including flux outside the FOV
        integral = ig.dblquad(
                self._sersic_func,
                -np.inf,
                np.inf,
                lambda y: -np.inf,
                lambda y: np.inf,
                args=(self.major, self.minor, self.sersic_index)
        )
        self.norm_prof = self.prof/integral[0] * self.pixelscale()
        
    def surface_scale(self):
        self.norm_prof = self.prof * np.e * self.pixelscale()

    def surface_center(self):
        self.norm_prof = self.prof * self.pixelscale()

    def _sersic_func(self, y, x, major, minor, index):
        """
        Implement Sersic intensity profile in a way that scipy.integrate.dblquad can use 
        since it is passed there as a callable. We need to integrate this function to 
        normalize it properly for flux outside the calculation FOV.

        Parameters
        ----------
        y: float or numpy.ndarray
            Y values for evaluating function
        x: float or numpy.ndarray
            X values for evaluating function
        major: float
            Major axis scale length
        minor: float
            Minor axis scale length
        index: float
            Sersic index

        Returns
        -------
        profile: float or numpy.ndarray
            Float or array containing evaluated Sersic profile
        """
        dist = np.sqrt((x / major)**2.0 + (y / minor)**2.0)
        profile = np.exp(-dist**(1.0 / index))
        return profile

class Gaussian2dDistribution(SersicDistribution):
    """
    The Gaussian 2D profile is a special case of the Sersic profile, where sersic index = 0.5.

    Parameters
    ----------
    src: dict
        Details of the source geometry to generate, from the Source block of the input files.

    Attributes
    ----------
    minor: float
        Minor axis scale length
    major: float
        Major axis scale length
    pa: float
        Position angle in degrees of major axis measured positive in +X direction
    xoff: float
        Offset in X direction
    yoff: float
        Offset in Y direction
    prof: np.ndarray
        raw 2D profile in detector units
    norm_prof: np.ndarray
        normalized 2D profile in detector units

    """
    def __init__(self,src):
        Distribution.__init__(self,src)
        self.major=src.shape['major']*np.sqrt(2.0) # to match the usual definition of a gaussian
        self.minor=src.shape['minor']*np.sqrt(2.0)
        self.sersic_index=0.5
        self.surf_area_units = src.shape['surf_area_units']

        
        self.prof = self._sersic_func(self.yrot, self.xrot, self.major, self.minor, self.sersic_index)
        
        self.normalize(src)

class FlatDistribution(Distribution):
    """
    Implement a source with a constant surface brightness as an ellipse, i.e. a tilted disc.
    
    Parameters
    ----------
    src: dict
        Details of the source geometry to generate, from the Source block of the input files.

    Attributes
    ----------
    minor: float
        Minor axis scale length
    major: float
        Major axis scale length
    pa: float
        Position angle in degrees of major axis measured positive in +X direction
    xoff: float
        Offset in X direction
    yoff: float
        Offset in Y direction
    prof: np.ndarray
        raw 2D profile in detector units
    norm_prof: np.ndarray
        normalized 2D profile in detector units

    Methods
    -------
        integrate_infinity():
            Normalization to flux at infinity means total flux, and that if all pixels in the 
            profile are summed up, they should sum to 1.
    surface_center():
        Normalization to the center is, in this case, a no-op, because the initial 
        definition of the disk sets the center (and everywhere else) to 1
    bright_effective():
        Normalization to average brightness within the effective radius (or here, radius)
        means that the intensity per square arcsecond should sum to 1. Thus, all we need 
        to take care of is the pixel scale in arcseconds^2
    """
    def __init__(self,src):
        Distribution.__init__(self,src)
        self.major=src.shape['major']
        self.minor=src.shape['minor']
        self.surf_area_units = src.shape['surf_area_units']


        self.prof = src.grid.elliptical_mask(self.major, self.minor, pa=self.pa, xoff=self.xoff, yoff=self.yoff)
        
        self.normalize(src)

    def integrate_infinity(self):
        self.norm_prof = self.prof / (np.pi * self.major * self.minor) * self.pixelscale()
        
    def surface_center(self):
        self.norm_prof = self.prof * self.pixelscale()




# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import division, absolute_import
import os

from .telescope import Telescope
from .instrument import Instrument
from .psf_library import PSFLibrary

class WFIRST(Telescope):

    """
    Currently a dummy class for directory/file discovery, but could eventually contain WFIRST-specific methods
    """
    pass


class WFIRSTInstrument(Instrument):

    """
    Generic WFIRST Instrument class
    """

    def __init__(self, mode=None, config={}, **kwargs):
        telescope = WFIRST()
        # these are the required sections and need to be passed via API in webapp mode
        self.instrument_pars = {}
        self.instrument_pars['detector'] = ["nexp", "ngroup", "nint", "readmode", "subarray"]
        self.instrument_pars['instrument'] = ["aperture", "disperser", "filter", "instrument", "mode"]
        self.api_parameters = list(self.instrument_pars.keys())

        # these are required for calculation, but ok to live with config file defaults
        self.api_ignore = ['dynamic_scene', 'max_scene_size', 'scene_size']

        Instrument.__init__(self, telescope=telescope, mode=mode, config=config, **kwargs)


class WFIRSTImager(WFIRSTInstrument):

    """
    Currently the WFIRSTImager requires only one methods beyond those provided by the generic Instrument class
    """
    def _loadpsfs(self):
        """
        Short-wavelength filters need PSFs that don't have the cold pupil mask (0.48-1.6 microns)
        Long-wavelength filters need PSFs that DO have the cold pupil mask (0.8-2.2 microns)
        Because the ranges overlap, we need to switch between PSF libraries.
        The mask is specifically baked into the filters, so selecting by filter seems appropriate
        """
        if self.instrument['filter'] in self.cold_pupil:
            psfs_key = "%s%s" % (self.instrument['aperture'], 'lw')
        else:
            psfs_key = "%s%s" % (self.instrument['aperture'], 'sw')

        psf_path = os.path.join(self.ref_dir, "psfs")
        self.psf_library = PSFLibrary(path=psf_path, aperture=psfs_key)


class WFIRSTIFU(WFIRSTInstrument):

    """
    Currently the WFIRSTIFU requires no extra methods beyond those provided by the generic Instrument class
    """
    pass



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import division, absolute_import

import numpy as np
from photutils.geometry import circular_overlap_grid, elliptical_overlap_grid

from .custom_exceptions import EngineInputError


class Grid(object):

    """
    Generate an x, y grid in a rectangular region, sampled with xsamp and
    ysamp spacings in the x and y directions, respectively.  Origin is at the
    upper-left corner to match numpy and matplotlib convention. astropy.io.fits
    also assumes UL origin by default.

    Parameters
    ----------
    xsamp, ysamp : float, float
        The sampling spacing in the x and y directions.
    nx, ny: int, int
        Number of samples in the x and y directions.
    """

    def __init__(self, xsamp, ysamp, nx, ny):
        self.xsamp = np.abs(xsamp)
        self.ysamp = np.abs(ysamp)
        startx = -(nx - 1) / 2.0 * self.xsamp
        stopx = (nx - 1) / 2.0 * self.xsamp
        starty = -(ny - 1) / 2.0 * self.ysamp
        stopy = (ny - 1) / 2.0 * self.ysamp
        xvals = np.linspace(startx, stopx, num=nx)
        yvals = np.linspace(starty, stopy, num=ny)

        ones = np.ones((ny, nx))
        x = ones * xvals
        y = np.flipud(ones * yvals.reshape(int(ny), 1))

        self.nx = nx
        self.ny = ny
        self.x = x
        self.y = y
        self.row = xvals
        # flip Y axis because we use Y increasing from bottom to top
        self.col = yvals[::-1]

    @property
    def shape(self):
        """
        Provides array-like .shape functionality
        """
        sh = (self.ny, self.nx)
        return sh

    def as_dict(self):
        """
        Create engine API format dict containing the properties of the Grid instance. The half-pixel
        offsets are due to the pixel coordinates being referenced to pixel centers. Thus the edges of the
        Grid are a half-pixel above and to the left of the Grid index origin.
        """
        transform = {}
        transform['x_refpix'] = 0
        transform['x_refval'] = self.x[0][0]
        transform['x_max'] = self.x.max() + 0.5 * self.xsamp
        transform['x_min'] = self.x.min() - 0.5 * self.xsamp
        transform['x_step'] = self.xsamp
        transform['x_size'] = self.x.shape[1]
        transform['y_refpix'] = 0
        transform['y_refval'] = self.y[0][0]
        transform['y_max'] = self.y.max() + 0.5 * self.ysamp
        transform['y_min'] = self.y.min() - 0.5 * self.ysamp
        transform['y_step'] = -self.ysamp
        transform['y_size'] = self.y.shape[0]
        return transform

    def wcs_info(self):
        """
        Define coordinate transform in WCS header format.

        Returns
        -------
        header: dict
            WCS keys defining coordinate transform for the 2 spatial axes
        """
        t = self.as_dict()
        header = {
            'ctype1': 'X offset',
            'crpix1': 1,
            'crval1': t['x_min'],
            'cdelt1': t['x_step'],
            'cunit1': 'arcsec',
            'cname1': 'X',
            'ctype2': 'Y offset',
            'crpix2': 1,
            'crval2': t['y_min'],
            'cdelt2': -t['y_step'],
            'cunit2': 'arcsec',
            'cname2': 'Y',
        }
        return header

    def __len__(self):
        """
        Provide array-like len() functionality
        """
        l = self.x.shape[0] * self.x.shape[1]
        return l

    def __getitem__(self, val):
        """
        Provide array-like indexing functionality.

        Parameters
        ----------
        val : slice object
            Valid python index or slice() specification.

        Returns
        -------
        pos : [y, x] list of floats or arrays
            Elements are floats in the case of specific indicies and arrays in
            the cases of slice()'s.
        """
        section = [self.y[val], self.x[val]]
        return section

    def bounds(self):
        return {'xmin': np.min(self.x),
                'xmax': np.max(self.x),
                'ymin': np.min(self.y),
                'ymax': np.max(self.y)}

    def dist(self, xcen=0.0, ycen=0.0):
        """
        Return a distance array where each element contains its distance from
        the center of the grid.
        """
        d = np.sqrt((self.x - xcen) ** 2 + (self.y - ycen) ** 2)
        return d

    def dist_xy(self, xcen=0.0, ycen=0.0):
        """
        Return a directional distance array where each element contains its (x,y) distance from
        the center of the grid.
        """
        dx = self.x - xcen
        dy = self.y - ycen
        return dx,dy

    def world_to_image(self, y, x):
        """
        Return fractional index coordinates yi,xi corresponding to given y,x position

        Parameters
        ----------
        y, x : float
            y, x position in world coordinates

        Returns
        -------
        pos : [yi, xi] list of floats
            yi, xi image coordinates that correspond to given position
        """
        t = self.as_dict()
        if x < t['x_min'] or x > t['x_max']:
            raise ValueError("X outside of Grid bounds.")
        if y < t['y_min'] or y > t['y_max']:
            raise ValueError("Y outside of Grid bounds.")
        x_image = x / self.xsamp + (self.shape[1] - 1) / 2.0
        # there's a minus sign here because y indices increase from up to down,
        # but y coordinates increase down to up.
        y_image = -y / self.ysamp + (self.shape[0] - 1) / 2.0
        pos = [y_image, x_image]
        return pos

    def world_to_index(self, y, x):
        """
        Return nearest index pair to given y,x position

        Parameters
        ----------
        y, x : float
            y,x position in world coordinates

        Returns
        -------
        pos : [y, x] list of ints
            y,x indices that most closely correspond to given position
        """
        y_index, x_index = self.world_to_image(y, x)
        y_index, x_index = self._index_bound(int(round(y_index)), self.ny), self._index_bound(int(round(x_index)), self.nx)
        pos = [y_index, x_index]
        return pos

    def _index_bound(self, index, nindex):
        """
        Apply bounds to indicies so the nearest index is either the first or last index for cases
        at edge or beyond Grid.

        Parameters
        ----------
        index: int
            Index to bounds-check
        nindex: int
            Length of axis being indexed

        Returns
        -------
        index: int
            Index with boundaries applied
        """
        if index < 0:
            index = 0
        if index > nindex - 1:
            index = nindex - 1
        return index

    def shift_rotate(self, yoff, xoff, rot):
        """
        Return shifted/rotated (y, x) given offsets (yoff, xoff) and rotation, rot (degrees)

        Parameters
        ----------
        yoff, xoff: float
            yoff, xoff offsets in world coordinates
        rot: float
            rotation angle in degrees

        Returns
        -------
        ysh_rot, xsh_rot: 2D numpy arrays
            rotated and shifted copies of Grid.x and Grid.y
        """
        pa_radians = np.pi * rot / 180.0
        xsh = self.x - xoff
        ysh = self.y - yoff
        xsh_rot = xsh * np.cos(pa_radians) + ysh * np.sin(pa_radians)
        ysh_rot = -xsh * np.sin(pa_radians) + ysh * np.cos(pa_radians)
        return ysh_rot, xsh_rot

    def get_aperture(self):
        """
        Return Grid parameters as an aperture specification
        """
        ap = {'width': self.nx * self.xsamp, 'height': self.ny * self.ysamp, 'offset': (0, 0)}
        return ap

    def rectangular_mask(self, width=None, height=None, xoff=0.0, yoff=0.0, transparency=1.0):
        """
        Define a rectangular mask within the Grid of specfied width and height and offset by xoff/yoff.
        Pixels partially covered by mask are properly weighted by the area subtended by the mask.

        Parameters
        ----------
        width: float
            Width of the mask. If None, then use the full width of the Grid.
        height: float
            Height of the mask. If None, then use the full height of the Grid.
        xoff: float
            X offset of the mask
        yoff: float
            Y offset of the mask
        transparency: float (default: 1.0)
            Transparency of the mask

        Returns
        -------
        mask: 2D np.ndarray
            2D mask image
        """
        if transparency < 0.0 or transparency > 1.0:
            msg = "Mask transparency, %f, must be in the range of 0.0 (fully opaque) to 1.0 (fully clear)." % transparency
            raise EngineInputError(value=msg)

        minx = self.row.min()
        maxx = self.row.max()
        xcen = (minx + maxx) / 2.0
        miny = self.col.min()
        maxy = self.col.max()
        ycen = (miny + maxy) / 2.0

        # if width and/or height are not specified, set them encompass the size of the Grid.
        # add two pixels to make sure the masking does not interpolate the edges.
        if width is None:
            width = maxx - minx + 2 * self.xsamp
        if height is None:
            height = maxy - miny + 2 * self.xsamp

        mask_x = np.array(
            [
                minx,
                xcen - (width + self.xsamp) / 2.0 + xoff,
                xcen - (width - self.xsamp) / 2.0 + xoff,
                xcen + (width - self.xsamp) / 2.0 + xoff,
                xcen + (width + self.xsamp) / 2.0 + xoff,
                maxx
            ]
        )
        mask_y = np.array(
            [
                miny,
                ycen - (height + self.ysamp) / 2.0 + yoff,
                ycen - (height - self.ysamp) / 2.0 + yoff,
                ycen + (height - self.ysamp) / 2.0 + yoff,
                ycen + (height + self.ysamp) / 2.0 + yoff,
                maxy
            ]
        )

        mask_pix = np.array([0.0, 0.0, 1.0, 1.0, 0.0, 0.0])
        x_mask_int = np.interp(self.row, mask_x, mask_pix)
        y_mask_int = np.interp(self.col, mask_y, mask_pix)
        x_mask = x_mask_int.reshape((1, self.nx))
        y_mask = y_mask_int.reshape((self.ny, 1))
        mask = y_mask * x_mask * transparency
        return mask

    def circular_mask(self, radius, xoff=0.0, yoff=0.0, use_exact=1, subsampling=1, transparency=1.0):
        """
        Define a circular mask within the grid of a specified radius and offset by xoff/yoff.
        Use photutils.geometry to create masks that calculate the fraction of a pixel subtended
        by the circular mask.  use_exact=1 performs the exact geometric calculation while
        use_exact=0 will sub-sample the pixels by the specfied subsampling parameter
        to estimate the subtended area.

        Parameters
        ----------
        radius: float
            Radius of the circular mask
        xoff: float
            X position of the center of the mask
        yoff: float
            Y position of the center of the mask
        use_exact: int (default: 1)
            If 1, then use exact geometrical calculation to weight pixels partially covered by mask.
            This parameter is passed directly on to photutils.geometry.circular_overlap_grid()
        subsampling: int (default: 1)
            If use_exact=0, then subsample pixels by this factor to calculate area of each pixel subtended
            by the mask. The default value of 1 will force no partial pixels to be included in the mask.
            This parameter is passed directly on to photutils.geometry.circular_overlap_grid()
        transparency: float (default: 1.0)
            Transparency of the mask
        Returns
        -------
        mask: 2D np.ndarray
            2D mask image
        """
        if transparency < 0.0 or transparency > 1.0:
            msg = "Mask transparency, %f, must be in the range of 0.0 (fully opaque) to 1.0 (fully clear)." % transparency
            raise EngineInputError(value=msg)

        t = self.as_dict()
        # we need to use flipud because we use an origin in the UL corner of an image
        # while photutils uses the LL corner.
        mask = np.flipud(
            circular_overlap_grid(
                t['x_min'] - xoff,
                t['x_max'] - xoff,
                t['y_min'] - yoff,
                t['y_max'] - yoff,
                t['x_size'],
                t['y_size'],
                radius,
                use_exact,
                subsampling
            )
        )
        mask *= transparency
        return mask

    def elliptical_mask(self, major, minor, pa=0.0, xoff=0.0, yoff=0.0, use_exact=1, subsampling=1, transparency=1.0):
        """
        Define an elliptical mask within the grid with specified major/minor axes, position angle,
        and offset from center by xoff/yoff. Use photutils.geometry to create masks that calculate
        the fraction of a pixel subtended by the mask.  use_exact=1 performs the exact geometric
        calculation while use_exact=0 will sub-sample the pixels by the specfied subsampling parameter
        to estimate the subtended area.

        Parameters
        ----------
        major: float
            Semi-major axis of the elliptical mask
        minor: float
            Semi-minor axis of the mask
        pa: float
            Position angle of the ellipse in degrees measured clockwise (positive in +X direction)
        xoff: float
            X position of the center of the mask
        yoff: float
            Y position of the center of the mask
        use_exact: int (default: 1)
            If 1, then use exact geometrical calculation to weight pixels partially covered by mask.
            This parameter is passed directly on to photutils.geometry.circular_overlap_grid()
        subsampling: int (default: 1)
            If use_exact=0, then subsample pixels by this factor to calculate area of each pixel subtended
            by the mask. The default value of 1 will force no partial pixels to be included in the mask.
            This parameter is passed directly on to photutils.geometry.circular_overlap_grid()
        transparency: float (default: 1.0)
            Transparency of the mask

        Returns
        -------
        mask: 2D np.ndarray
            2D mask image
        """
        if transparency < 0.0 or transparency > 1.0:
            msg = "Mask transparency, %f, must be in the range of 0.0 (fully opaque) to 1.0 (fully clear)." % transparency
            raise EngineInputError(value=msg)

        theta = np.pi * pa / 180.0  # photutils uses angles in radians

        t = self.as_dict()
        # we need to use flipud because we use an origin in the UL corner of an image
        # while photutils uses the LL corner.
        mask = np.flipud(
            elliptical_overlap_grid(
                t['x_min'] - xoff,
                t['x_max'] - xoff,
                t['y_min'] - yoff,
                t['y_max'] - yoff,
                t['x_size'],
                t['y_size'],
                major,
                minor,
                theta,
                use_exact,
                subsampling
            )
        )
        mask *= transparency
        return mask

    def point_source(self, xoff=0.0, yoff=0.0):
        """
        Define a point source within the grid at position xoff, yoff.  Sub-pixel positioning
        of the point source is performed by weighting the 4 nearest pixels.

        Parameters
        ----------
        xoff: float
            X position of point source
        yoff: float
            Y position of point source

        Returns
        -------
        im: 2D np.ndarray
            Image containing point source with flux scaled to 1.0.
        """
        im = np.zeros(self.shape)

        # normally it's an error to specify a position off the Grid, but in this case it just
        # means there's no flux being added onto the image. in cases such as dithering, putting
        # the point source off the Grid is intentional and shouldn't generate errors or warnings.
        try:
            yi, xi = self.world_to_image(yoff, xoff)
        except ValueError as e:
            return im

        rem_x, int_x = np.modf(xi)
        rem_y, int_y = np.modf(yi)

        # check for image edges and handle weights correctly if point source straddles an edge
        if rem_x < 0.0:
            x_shift = -1
            rem_x = 1.0 + rem_x
        else:
            x_shift = 0

        if rem_y < 0.0:
            y_shift = -1
            rem_y = 1.0 + rem_y
        else:
            y_shift = 0

        # here we handle the cases where a point source is at the edge of the Grid so that some of the
        # flux is in pixels on the Grid, while the rest is off the Grid.
        ly = int(int_y + y_shift)
        uy = int(ly + 1)
        lx = int(int_x + x_shift)
        ux = int(lx + 1)

        # if a pixel is None, then it's off the Grid and we don't try to set it.
        if ly < 0:
            ly = None
        if uy == self.ny:
            uy = None
        if lx < 0:
            lx = None
        if ux == self.nx:
            ux = None

        if ly is not None and lx is not None:
            im[ly, lx] += (1.0 - rem_x) * (1.0 - rem_y)
        if uy is not None and lx is not None:
            im[uy, lx] += (1.0 - rem_x) * rem_y
        if ly is not None and ux is not None:
            im[ly, ux] += rem_x * (1.0 - rem_y)
        if uy is not None and ux is not None:
            im[uy, ux] += rem_x * rem_y

        return im


class IrregularGrid(Grid):

    """
    Create Grid from pre-specified rows and cols.
    """

    def __init__(self, yvals, xvals):
        ny = yvals.size
        nx = xvals.size
        ones = np.ones((ny, nx))
        y = ones * yvals.reshape(ny, 1)
        x = ones * xvals

        self.nx = nx
        self.ny = ny
        self.x = x
        self.y = y
        self.row = xvals
        self.col = yvals
        if len(xvals) > 1:
            self.xsamp = np.abs(xvals[1] - xvals[0])
        if len(yvals) > 1:
            self.ysamp = np.abs(yvals[1] - yvals[0])



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import division, absolute_import

import six
from functools import reduce

import numpy as np
import pysynphot as psyn

from .custom_exceptions import EngineInputError, PysynphotError

default_separator = "__"


def merge_wavelengths(waveset1, waveset2, threshold=1.0e-12):
    """
    Return the union of the two sets of wavelengths using
    :func:`numpy.union1d`.  This is cribbed from astropy.synphot.utils;
    the original version was in astrolib pysynphot.spectrum.MergeWavesets.

    The merged wavelengths may sometimes contain numbers which are nearly
    equal but differ at levels as small as 1e-14. Having values this
    close together can cause problems down the line. So, here we test
    whether any such small differences are present, with a small
    difference defined as less than ``threshold``. If a small
    difference is present, the lower of the too-close pair is removed.

    Parameters
    ----------
    waveset1, waveset2 : array_like
        Wavelength values, assumed to be in the same unit already.

    threshold : float, optional
        Merged wavelength values are considered "too close together"
        when the difference is smaller than this number.
        The default is 1e-12.

    Returns
    -------
    out_wavelengths : array_like
        Merged wavelengths.

    """
    out_wavelengths = np.union1d(waveset1, waveset2)
    delta = out_wavelengths[1:] - out_wavelengths[:-1]
    i_good = np.where(delta > threshold)

    # Remove "too close together" duplicates
    if len(i_good[0]) < delta.size:
        out_wavelengths = np.append(out_wavelengths[i_good],
                                    out_wavelengths[-1])
    return out_wavelengths


def spectrum_resample(flux, orig_wave, new_wave, mask_val=np.nan):
    """
    Use pysynphot to re-sample a spectrum to a new set of wavelengths while conserving flux.

    Parameters
    ----------
    flux: 1D np.ndarray
        Input spectrum to be re-binned
    orig_wave: 1D np.ndarray
        Set of wavelengths for input spectrum
    new_wave: 1D np.ndarray
        New set of wavelengths to re-bin spectrum onto
    mask_val: float or np.nan (default: np.nan)
        Value to fill in where new_wave is outside the bounds of orig_wave. np.nan is the right thing
        to use here, but make it configurable in case it needs to be changed.

    Returns
    -------
    binned_flux: 1D np.ndarray
        Input spectrum re-binned onto new_wave
    """
    # if wavelength sets are the same, then pass back input flux unmodified.  the input wavelengths set the nyquist limit
    # so if you resample a spectrum back onto it's own wavelength set, it still gets downgraded if it was undersampled.
    if (orig_wave.size == new_wave.size) and np.allclose(orig_wave, new_wave):
        binned_flux = flux
    else:
        try:
            spec = psyn.spectrum.ArraySourceSpectrum(wave=orig_wave, flux=flux)
            f = np.ones(len(orig_wave))
            filt = psyn.spectrum.ArraySpectralElement(orig_wave, f, waveunits='microns')
            obs = psyn.observation.Observation(spec, filt, binset=new_wave, force='extrap')
            binned_flux = obs.binflux
        except Exception as e:
            msg = "Error resampling spectrum to new waveset via Pysynphot: %s" % type(e)
            raise PysynphotError(value=msg)

        # the use of 'extrap' means pysynphot will fill in the last available flux value for any
        # wavelengths beyond the bounds of the original spectrum. this is going to be wrong for just
        # about any circumstance. the right thing to do is to recognize that we don't know what's beyond
        # the bounds of the original spectrum and use np.nan as the fill value for fluxes at these new
        # wavelengths. mask_val defaults to np.nan, but is configurable in case there's a need to use
        # something else.
        wmin = orig_wave.min()
        wmax = orig_wave.max()
        invalid_subs = np.where((new_wave < wmin) | (new_wave > wmax))
        binned_flux[invalid_subs] = mask_val

    return binned_flux


def recursive_subclasses(cls):
    """
    The __subclasses__() method only goes on level deep, but various classes that ultimately
    inherit from things like Instrument and Strategy are separated by multiple inheritance layers.
    This function recursively walks through the inheritance tree and returns a list of all subclasses
    at all levels that inherit from the given class.

    Parameters
    ----------
    cls: any python Class
        Python class to get list of subclasses for

    Returns
    -------
    all_subclasses: list
        List of all subclasses that ultimately inherit from cls
    """
    all_subclasses = []

    top_subclasses = cls.__subclasses__()
    all_subclasses.extend(top_subclasses)

    for s in top_subclasses:
        all_subclasses.extend(recursive_subclasses(s))

    return all_subclasses


def merge_data(*dicts):
    """
    This takes a list of python dicts and merges them into a single dict.  It is set up to
    assure that later arguments will take precedence over earlier ones by default.

    Parameters
    ----------
    dicts: list
        List of dicts to merge into a single dict

    Returns
    -------
    updated: dict
        Arguments combined into a single dict.  Later arguments take precedence over earlier arguments
        and dicts take precedence over non-dict values.
    """
    updated = {}

    # grab all of the keys
    keys = set()
    for o in dicts:
        keys = keys.union(set(o))

    for key in keys:
        values = [o[key] for o in dicts if key in o]
        # find values that are dicts so we can recurse through them
        maps = [value for value in values if isinstance(value, dict)]
        if maps:
            updated[key] = merge_data(*maps)
        else:
            # if not a dict, then return the last value we have since later arguments
            # take precendence
            updated[key] = values[-1]
    return updated


def get_key_list(key, separator=default_separator):
    """
    Create list of keys from a flattened key to set/retrieve value from nested dict
    """
    keylist = key.split(separator)
    return keylist


def flat_key_from_list(list, separator=default_separator):
    """
    Create a flattened version of a nested key from a key list
    """
    flat_key = separator.join(list)
    return flat_key


def get_dict_from_keys(data, keylist):
    """
    Use a list of keys to recurse through a dict to get a value

    >>> d = {'a': {'b': 2}}
    >>> get_dict_from_keys(d, ['a', 'b'])
    2
    """
    try:
        val = reduce(dict.__getitem__, keylist, data)
    except KeyError as e:
        msg = "Invalid keys provided: %s (%s)" % (repr(keylist), e)
        raise EngineInputError(value=msg)
    return val


def set_dict_from_keys(data, keylist, value):
    """
    Use a list of keys to recurse through a dict to set a value

    >>> d = {'a': {'b': 2}}
    >>> set_dict_from_keys(d, ['a', 'b'], 6)
    >>> print d
    {'a': {'b': 6}}
    """
    get_dict_from_keys(data, keylist[:-1])[keylist[-1]] = value


def lower_key(in_obj):
    """
    take input object and recursively make keys and strings lower-case and convert
    spaces to underlines to ensure compliance with pyetc coding standards.

    Parameters
    ----------
    in_obj: python object
        object containing keys and values to be modified. if it is not a dict, list, str, or unicode
        it is simply returned to the caller unchanged.

    Returns
    -------
    obj: python object
        object of same type as in_obj with strings modifyed to be lower case and spaces removed
    """
    if isinstance(in_obj, dict):
        out_dict = {}
        for key, item in list(in_obj.items()):
            out_dict[key.lower().replace(' ', '_')] = lower_key(item)
        obj = out_dict
    elif isinstance(in_obj, list):
        obj = [lower_key(obj) for obj in in_obj]
    elif isinstance(in_obj, str) or isinstance(in_obj, six.text_type):
        obj = in_obj.lower().replace(' ', '_')
    else:
        obj = in_obj
    return obj



"""Functions for getting and saving SVN info for distribution."""


from __future__ import with_statement

import os
import subprocess

from .astutils import ImportVisitor, walk


def get_svn_version(path='.'):
    """Uses ``svnversion`` to get just the latest revision at the given
    path.
    """

    try:
        pipe = subprocess.Popen(['svnversion', path], stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
    except OSError:
        return None

    if pipe.wait() != 0:
        return None

    return pipe.stdout.read().decode('latin1').strip()


def get_svn_info(path='.'):
    """Uses ``svn info`` to get the full information about the working copy at
    the given path.
    """

    path = os.path.abspath(path)

    try:
        pipe = subprocess.Popen(['svn', 'info', path], stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
        # stderr is redirected in order to squelch it.  Later Python versions
        # have subprocess.DEVNULL for this purpose, but it's not available in
        # 2.5
    except OSError:
        return 'unknown'

    if pipe.wait() != 0:
        return 'unknown'

    lines = []
    for line in pipe.stdout.readlines():
        line = line.decode('latin1').strip()
        if not line:
            continue
        if line.startswith('Path:'):
            line = 'Path: %s' % os.path.basename(path)
        lines.append(line)

    if not lines:
        return 'unknown'

    return '\n'.join(lines)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import division, absolute_import

import os
import numpy as np
import pysynphot as psyn

from . import config as cf
from .config import DefaultConfig
from .io_utils import ref_data_column
from .utils import spectrum_resample
from .custom_exceptions import EngineInputError, DataError, PysynphotError
import six

default_refdata_directory = cf.default_refdata_directory


class Background(cf.DefaultConfig):

    """
    This class provides the background flux.  The background can be computed from the engine's
    notional reference data or passed as a spectrum of a form [wave, sb] where wave is an array of
    wavelengths in microns and sb is an array of surface brightnesses in MJy/sr.

    Parameters
    ----------
    observation: Observation instance
        Contains needed instrument and background configuration data.
    config: dict
        Extra configuration information
    webapp: bool
        Enable strict API checking (not yet implemented here)

    Attributes
    ----------
    ref_dir: string
        Directory containing data for notional background model
    bg_level: string or list-like or array-like
        Input background level. Can be string for configuring notional model or lists/arrays for user-input background.
    instrument: Instrument instance
        Instrument configuration data is required to determine pixel scale and telescope/instrument contributions
        to the background model.
    aperture_pars: dict
        Instrument configuration data containing pixel scale required to calculate per-pixel background flux
    bg_spec: pysynphot.ArraySpectrum
        Original full-length, full-resolution background spectrum with wavelengths in microns and surface brightness in MJy/sr.
        pysynphot doesn't natively support mega-janskys so fake it for now by using fluxunits='jy' so that it at least knows
        the flux density is in F_nu.
    wave: 1D np.ndarray
        Binned/trimmed wavelength set
    MJy_sr: 1D np.ndarray
        Binned/trimmed background spectrum in units of MJy/sr

    Properties
    ----------
    mjy_pix: 1D np.ndarray
        Convert self.MJy_sr to units of mJy (milli-janskys) per pixel
    flambda_pix: 1D np.ndarray
        Convert self.MJy_sr to units of F_lambda per pixel, photons/cm^2/s/micron/pixel.
    """

    def __init__(self, observation, config={}, webapp=False, **kwargs):
        DefaultConfig.__init__(self, config=config, webapp=webapp, **kwargs)
        self.webapp = webapp
        self.ref_dir = os.path.join(default_refdata_directory, 'background')
        self.bg_level = observation.background
        self.instrument = observation.instrument
        self.aperture_pars = self.instrument.get_aperture_pars()
        # if the background spectrum is provided, the 0th index is wavelength and the 1st is the surface brightness
        if isinstance(self.bg_level, (list, tuple, np.ndarray)):
            try:
                # wavelength set for background that will get resampled to match wavelength set for a scene
                self.wave = np.array(self.bg_level[0], dtype=np.float64)
                # background spectrum in standard units of MJy/sr. use properties mjy_pix and flambda_pix to convert units.
                self.MJy_sr = np.array(self.bg_level[1], dtype=np.float64)
                # we store the original bg spectrum in a pysynphot spectrum so that resampling operations
                # are done on the original data.
                self.bg_spec = psyn.ArraySpectrum(wave=self.wave, flux=self.MJy_sr, waveunits='microns', fluxunits='jy')
            except Exception as e:
                msg = "Malformed input background spectrum: %s (%s)" % (repr(self.bg_level), e)
                raise EngineInputError(value=msg)
        elif isinstance(self.bg_level, (str, six.text_type)):
            self.notional_background()
        else:
            msg = "Must provide a valid background spectrum or background signifier string: %s" % repr(self.bg_level)
            raise EngineInputError(value=msg)

    def resample(self, wavelengths):
        """
        Re-bin background onto a new set of wavelengths.

        This method modifies self.wave and self.MJy_sr, but leaves self.bg_spec alone.

        Parameters
        ----------
        wavelengths: 1D numpy array
            Array of new wavelengths to sample spectrum onto
        """
        self.MJy_sr = spectrum_resample(self.bg_spec.flux, self.bg_spec.wave, wavelengths)
        self.wave = wavelengths

    def telescope(self):
        """
        Load the telescope background.

        Returns
        -------
        bg_spec : pysynphot.ArraySpectrum
            Telescope background in MJy/sr
        """
        try:
            bg_tel_wave, bg_tel = self.instrument.telescope.get_bg_tel()
            bg_spec = psyn.ArraySpectrum(wave=bg_tel_wave, flux=bg_tel, waveunits='microns')
        except Exception as e:
            msg = "Problem loading telescope background model. "
            if self.webapp:
                msg += "(%s)" % type(e)
            else:
                msg += repr(e)
            raise DataError(value=msg)
        return bg_spec

    def zodi(self):
        """
        Load the zodiacal background

        Returns
        -------
        bg_spec : pysynphot.ArraySpectrum
            Zodiacal background in MJy/sr
        """
        reffile = "zodi_" + self.bg_level + '.fits'
        bg_file = os.path.join(self.ref_dir, reffile)
        try:
            bg_wave = ref_data_column(bg_file, colname='WAVELENGTH')
            bg = ref_data_column(bg_file, colname='SB')
            bg_spec = psyn.ArraySpectrum(wave=bg_wave, flux=bg, waveunits='microns')
        except Exception as e:
            msg = "Problem loading zodical background model. "
            if self.webapp:
                msg += "(%s)" % type(e)
            else:
                msg += repr(e)
            raise DataError(value=msg)
        return bg_spec

    def cirrus(self):
        """
        Load the galactic IR background

        Returns
        -------
        bg_spec : pysynphot.ArraySpectrum
            Galactic IR background in MJy/sr
        """
        reffile = "cirrus_" + self.bg_level + '.fits'
        bg_file = os.path.join(self.ref_dir, reffile)
        try:
            bg_wave = ref_data_column(bg_file, colname='WAVELENGTH')
            bg = ref_data_column(bg_file, colname='SB')
            bg_spec = psyn.ArraySpectrum(wave=bg_wave, flux=bg, waveunits='microns', fluxunits='jy')
        except Exception as e:
            msg = "Problem loading galactic IR background model. "
            if self.webapp:
                msg += "(%s)" % type(e)
            else:
                msg += repr(e)
            raise DataError(value=msg)
        return bg_spec

    def notional_background(self):
        """
        Read reference data for pandeia's notional background model and configure self.wave, self.bg_level, and self.bg_spec to
        use it. This is based on the legacy background data used by scamp and will eventually be replaced by real BMG
        implementation.

        """
        # load the telescope background
        tel_bg = self.telescope()

        # load the cirrus and zodiacal backgrounds.  which ones to use will be keyed off of
        # self.bg_level.
        if self.bg_level in ('low', 'medium', 'high'):
            zodi_bg = self.zodi()
            cirrus_bg = self.cirrus()
            tot_bg = tel_bg + zodi_bg + cirrus_bg
        elif self.bg_level == 'none' or self.bg_level is None:
            bg_wave = np.array([0.1, 30.0])
            bg_sb = np.array([0.0, 0.0])
            try:
                tot_bg = psyn.ArraySpectrum(wave=bg_wave, flux=bg_sb, waveunits='microns', fluxunits='jy')
            except Exception as e:
                msg = "Problem creating background spectrum with Pysynphot. "
                if self.webapp:
                    msg += "(%s)" % type(e)
                else:
                    msg += repr(e)
                raise PysynphotError(value=msg)
        else:
            msg = "Invalid specification for notional background model: %s" % self.bg_level
            raise EngineInputError(value=msg)
        self.bg_spec = tot_bg
        self.wave = tot_bg.wave
        self.MJy_sr = tot_bg.flux

    @property
    def mjy_pix(self):
        """
        Convert background surface brightness in MJy/sr to F_nu per pixel in
        mJy/pixel (self.wave assumed to be in microns)

        Returns
        -------
        i_pix: 1D np.ndarray
            Flux per pixel in mJy/pixel
        """
        i_nu = 1e9 * self.MJy_sr * 2.3504e-11  # 1e9 convert MJy to mJy; 2.3504e-11 converts sr to square arcsec

        i_nu_pix = i_nu * self.aperture_pars['pix'] ** 2
        return i_nu_pix

    @property
    def flambda_pix(self):
        """
        Convert background surface brightness in MJy/sr to F_lambda per pixel in
        photons/cm^2/s/micron/pixel (self.wave assumed to be in microns)

        Returns
        -------
        i_pix: 1D np.ndarray
            Flux per pixel in photons/s/cm^2/pixel
        """
        i_lambda = 1.50916e9 / self.wave * self.MJy_sr * 2.3504e-11
        i_lambda_pix = i_lambda * self.aperture_pars['pix'] ** 2
        return i_lambda_pix



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
io_utils - commonly used I/O routines for JSON, FITS, and possibly other formats.
"""

from __future__ import division, absolute_import

import os
import json
import errno
import numpy as np
import astropy.io.fits as fits
import pysynphot as psyn

from .custom_exceptions import EngineInputError, DataError
from .constants import PANDEIA_WAVEUNITS, PANDEIA_FLUXUNITS


class NumPyArangeEncoder(json.JSONEncoder):

    """
    custom encoder to handle numpy arrays that might show up in inputs and outputs:

    http://stackoverflow.com/questions/11561932/why-does-json-dumpslistnp-arange5-fail-while-json-dumpsnp-arange5-tolis

    Parameters
    ----------
    obj: python object
        currently only supports translating np.ndarray and np.float32 objects

    Returns
    -------
    e: json.JSONEncoder object
        JSONEncoder that now understands the custom data types
    """

    def default(self, obj):
        # turn ndarrays into lists
        if isinstance(obj, np.ndarray):
            return obj.tolist()  # or map(int, obj)
        # numpy.float32's started showing up when refactoring refdata FITS files
        # catch them and make them into normal float()'s
        if isinstance(obj, np.float32):
            return float(obj)
        e = json.JSONEncoder.default(self, obj)
        return e


def get_json_parameters(parameters, field, instrument=None, detector=None, aperture=None, subarray=None, readmode=None, instmode=None):
    """
    This was primarly written (or initially written) for the detector parameters
    to read in the structure so values can be dependent on aperture, subarray
    readmode and/or instrument mode.  An example would be:

        "det_pars": {
        "det_type": "h2rg",
        "dark_current": 0.00636,
        "excessp1": 0.0,
        "excessp2": 0.0,
        "ff_electrons": 62500,
        "fullwell": {
			"default": 65000.0,
			"key_format": ["subarray", "readmode"],
			"full__nrsrapid": 30000.0,
			"s200a1__nrs": 22000.0
		},
        "ipc": true,
        "mingroups": 2,
        "pix_size": 18.0,
        "rn": {
			"default": 16.8,
			"key_format": ["aperture"],
			"sw": 17
		},
        "rn_correlation": true
    }

    This could be made more general (e.g., kwargs for the parameters) but for now
    going to leave as-is.

    :param field:
    :param aperture:
    :param subarray:
    :param readmode:
    :param instmode:
    :return:
    """

    if instrument and detector:
        aperture = instrument['aperture']
        subarray = detector['subarray']
        readmode = detector['readmode']
        instmode = instrument['mode']

    if field in parameters and not isinstance(parameters[field], dict):
        # If the field exists and is not a dict then just return the value
        return parameters[field]

    elif field in parameters and isinstance(parameters, dict):
        # If the field exists and it is a dict then we have to check the key_format
        # In this case the format of the parameter will be something like:
        # "rn": {
        #     "default": 16.8,
        #     "key_format": ["aperature"],
        #     "sw": 17
        # },

        # Create the key lookup
        key_lookup = []
        for kf in parameters[field]["key_format"]:
            if kf == "aperture":
                key_lookup.append(aperture)
            elif kf == "subarray":
                key_lookup.append(subarray)
            elif kf == "readmode":
                key_lookup.append(readmode)
            elif kf == "instmode":
                key_lookup.append(instmode)
            else:
                raise DataError("Unknown key format {} listed in {}".format(kf, parameters[field]))

        # Now squish the list to a "__" separated key and see if it exists
        # If all required keys were passed in then create teh string lookup
        # and then if that key is present in the dict return it.
        if not any([x is None for x in key_lookup]):
            key_lookup = '__'.join(key_lookup)

            if key_lookup in parameters[field]:
                return parameters[field][key_lookup]

        # If they key_lookup was not present or some of the required
        # data was not present then just return the default.
        if "default" in parameters[field]:
            return parameters[field]["default"]
        else:
            raise DataError("There is no default value in {}".format(parameters[field]))

    else:
        raise Exception("Bad key/value {} for det_pars {}".format(field, parameters))


def get_full_refwave_dict(data_dir, telescope, inst_list, modename_list):
    """
    Traverse the pandeia_data tree for the instruments and modenames given
    and find all refwave info (min, max, midpoint) that there is and return it in
    a tree of dicts.

    Parameters
    ----------
    data_dir: str
        Path to pandeia_data dir (e.g. $pandeia_refdata value)
    telescope: str
        E.g. 'jwst'
    inst_list: list of str
        instrument names for data desired, e.g. ['miri','nircam','niriss','nirspec'],
    modename_list: list of str
        mode names desired for all instruments listed above, e.g. ['imaging','mrs','soss','wfgrism']
    """
    retval = {}
    for inst in inst_list:
        this_inst = {}
        inst_cfg_fname = data_dir+os.sep+telescope+os.sep+inst+os.sep+'config.json'
        assert os.path.exists(inst_cfg_fname), "Error: cfg file not found: "+inst_cfg_fname
        inst_cfg = read_json(inst_cfg_fname, raise_except=True)
        all_inst_modes = inst_cfg['modes']
        mode_cfg = inst_cfg['mode_config']
        range_cfg = inst_cfg['range']
        assert sorted(all_inst_modes) == sorted(mode_cfg.keys()) or (inst == 'nirspec' and 1+len(all_inst_modes) == len(mode_cfg.keys())), "Error: modes list mismatch for: "+inst
        requested_modes = [m for m in modename_list if m in all_inst_modes]
        for mode in requested_modes:
            this_mode = {}
            inst_mode_apers = mode_cfg[mode]['apertures']
            for aper in inst_mode_apers:
                this_aper = {}
                if aper in range_cfg:
                    # we have matching range data !
                    for item in range_cfg[aper]:
                        if isinstance(range_cfg[aper][item], dict) and 'wmin' in range_cfg[aper][item] and 'wmax' in range_cfg[aper][item]:
                            this_item = range_cfg[aper][item].copy() # use a copy for below
                            mid = (this_item['wmin'] + this_item['wmax'])/2.0
                            mid = (int(100*mid))/100. # keep only to 2nd decimal place
                            this_item['wmidpoint'] = mid
                            this_aper[item] = this_item
                # done with aper
                if len(this_aper.keys()) > 0:
                    this_mode[aper] = this_aper
            # done with mode
            if len(this_mode.keys()) > 0:
                this_inst[mode] = this_mode
        # done with inst
        if len(this_inst.keys()) > 0:
            retval[inst] = this_inst
    return retval


def read_json(filename, raise_except=False, **kwargs):
    """
    read in a JSON format file.  return None if the file is not there.

    Parameters
    ----------
    filename: string
        name of the JSON file to read
    except: bool
        if true, raise exception if file is missing. if false, return empty dict.

    Returns
    -------
    d: python object
        data from JSON file decoded into python object
    """
    try:
        with open(filename, 'r') as f:
            json_data = json.load(f, **kwargs)
    except IOError as e:
        if e.errno == errno.ENOENT and raise_except is False:  # No such file
            json_data = {}
        else:
            msg = "Missing JSON file: %s" % filename
            raise EngineInputError(value=msg)
    d = json_data
    return d


def write_json(data, filename, **kwargs):
    """
    write python object into a JSON format file

    Parameters
    ----------
    data: python object
        python object to encode and write to JSON
    filename: string
        name of file to write JSON-encoded data to
    """
    with open(filename, 'w') as f:
        json.dump(data, f, indent=4, cls=NumPyArangeEncoder, separators=(',', ': '), **kwargs)
        f.write('\n')


def append_json(obj, filename):
    """
    read data structure from JSON, append it to a python object, and return the result.
    currently only supported for lists and dicts

    Parameters
    ----------
    obj: python object
        object to be appended to
    filename: string
        filename containing JSON data to be appended

    Returns
    -------
    obj: python object
        input python object now updated with appended data
    """
    json_data = read_json(filename)
    if isinstance(obj, list):
        if isinstance(json_data, list):
            for k in json_data:
                obj.append(k)
        else:
            obj.append(json_data)
    elif isinstance(obj, dict):
        if isinstance(json_data, dict):
            obj.update(json_data)
    else:
        raise ValueError("Can only append to list or dict.")
    return obj


def ref_data_interp(filename, wave, colname=None):
    """
    Read reference data from a FITS file and interpolate it to a provided wavelength array

    Parameters
    ----------
    filename: str
        Filename of reference data
    wave: numpy.ndarray
        Wavelength vector that the reference data will be interpolated onto
    colname: str
        Name of column within the reference file to read and interpolate

    Returns
    -------
    interp_col: numpy.ndarray
        Vector containing reference data interpolated onto wave
    """
    if colname is None:
        raise EngineInputError(value="Must specify name of column to read from reference file.")
    try:
        data = fits.getdata(filename)
    except IOError as e:
        error_msg = "Error reading reference file: " + filename
        raise DataError(value=error_msg)
    if np.any(np.diff(data['wavelength']) < 0):
        indices = np.where(np.diff(data['wavelength']) < 0)[0]
        error_msg = "Wavelengths must be increasing in reference file: %s\n" % (filename)
        error_msg += "Out-of-order indices: %s" % repr(indices)
        raise DataError(value=error_msg)
    try:
        columns = set(k.name.lower() for k in data.columns)
        if colname.lower() not in columns:
            msg = "Column %s not found in %s" % (colname, filename)
            raise DataError(value=msg)
        interp_col = np.interp(wave, data['wavelength'], data[colname])
    except Exception as e:
        error_msg = "Error interpolating reference file: %s : %s" % (filename, type(e))
        raise DataError(value=error_msg)
    return interp_col


def ref_data_column(filename, colname=None, error_msg="Error loading reference file."):
    """
    Read reference data from a FITS file and provide it directly with no interpolation

    Parameters
    ----------
    filename: str
        Filename of reference data
    colname: str
        Name of column within the reference file to read and return
    error_msg: str
        Custom error message to produce in case of an error

    Returns
    -------
    col: numpy.ndarray
        Vector containing requested reference data
    """
    if colname is None:
        raise EngineInputError(value="Must specify name of column to read from reference file.")
    try:
        data = fits.getdata(filename)
    except IOError as e:
        raise DataError(value="I/O " + error_msg)
    try:
        col = data[colname]
    except KeyError as e:
        raise DataError(value="Column not found in reference file: %s" % colname)
    return col


def read_psyn_spectrum(path):
    """
    Read spectrum in FITS or compliant ascii format (wavelength in angstroms, flux in flam).
    See http://etc.stsci.edu/etcstatic/users_guide/1_ref_5_user_spectra.html for details of
    acceptable files for pysynphot.

    Parameters
    ----------
    path: string
        Full pathname to FITS file containing spectrum

    Returns
    -------
    wave, flux:  1D np.ndarray, 1D np.ndarray
        Wavelength and flux vectors in 'PANDEIA_WAVEUNITS' and 'PANDEIA_FLUXUNITS', respectively
    """
    sp = psyn.FileSpectrum(path)
    sp.convert(PANDEIA_WAVEUNITS)
    sp.convert(PANDEIA_FLUXUNITS)
    wave, flux = sp.wave, sp.flux
    return wave, flux


def mkdir_p(path):
    """
    Implement 'mkdir -p' functionality with pure python

    Parameters
    ----------
    path: valid path specification
    """
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise
